from iotile.core.exceptions import *
from iotile.core.dev.iotileobj import IOTile
import os
import json
import resolvers
import re

class DependencyResolverChain (object):
	"""A set of rules mapping dependencies to DependencyResolver instances
	
	DependencyResolverChains let you customize which dependencies are looked up
	in which way.  For example, you could have some dependencies resolved using
	github while others must be installed locally.  

	The default DependencyResolverChain requires that all dependencies be installed
	locally and registered with the IOTile Registry.
	"""

	def __init__(self, settings_file=None):
		self.rules = []

		#FIXME: Load settings_file

		self.rules.append((re.compile('.*'), resolvers.ComponentRegistryResolver, {}))

	def update_dependency(self, tile, depinfo):
		"""Attempt to install or update a dependency to the latest version.

		Params
		======
		tile: An IOTile object describing the tile that has the dependency
		depinfo: a dictionary from tile.dependencies specifying the dependency
		
		Returns
		=======
		a string indicating the outcome.  Possible values are:
		"already installed"
		"installed"
		"updated"
		"not found"

		"""

		destdir = os.path.join(tile.folder, 'build', 'deps', depinfo['unique_id'])
		has_version = False
		if os.path.exists(destdir):
			has_version = True

		for rule in self.rules:
			if not self._check_rule(rule, depinfo):
				continue

			resolver = self._find_resolver(rule)

			if has_version:
				deptile = IOTile(destdir)

				#If the dependency is up to date, don't do anything
				if self._check_dep(depinfo, deptile, resolver):
					continue

				#Otherwise remove the current dependency and reload it
				import shutil
				shutil.rmtree(destdir)

			#Now try to resolve this dependency with the latest version
			result = resolver.resolve(depinfo, destdir)
			if result['found'] == False and result.get('stop', False):
				return 'not found'

			if result['found'] == False:
				continue

			settings = {
				'resolver': resolver.__class__.__name__
			}
			if 'settings' in result:
				settings['settings'] = result['settings']

			self._save_depsettings(destdir, settings)

			if has_version:
				return "updated"
			else:
				return "installed"

		if has_version:
			return "already installed"

		return "not found"


	def _save_depsettings(self, destdir, settings):
		settings_file = os.path.join(destdir, 'dep_settings.json')

		with open(settings_file, 'wb') as f:
			json.dump(settings, f, indent=4)

	def _load_depsettings(self, deptile):
		settings_file = os.path.join(deptile.folder, 'dep_settings.json')

		with open(settings_file, 'rb') as f:
			settings = json.load(f)

		return settings

	def _check_dep(self, depinfo, deptile, resolver):
		"""Check if a dependency tile is up to date
		"""

		try:
			settings = self._load_depsettings(deptile)
		except IOError as e:
			return False

		#If this dependency was initially resolved with a different resolver, then 
		#we cannot check if it is up to date
		if settings['resolver'] != resolver.__class__.__name__:
			return False

		resolver_settings = {}
		if 'settings' in settings:
			resolver_settings = settings['settings']

		return resolver.check(depinfo, deptile, resolver_settings)
	
	def _find_resolver(self, rule):
		regex,factory,args = rule
		return factory(args)

	def _check_rule(self, rule, depinfo):
		regex,factory,args = rule

		if regex.match(depinfo['name']):
			return True

		return False


from depresolver import DependencyResolver
from iotile.core.exceptions import *

class ComponentRegistryResolver (DependencyResolver):
	def __init__(self, settings={}):
		pass

	def resolve(self, depinfo, destdir):
		from iotile.core.dev.registry import ComponentRegistry

		reg = ComponentRegistry()

		try:
			comp = reg.find_component(depinfo['name'])
		except ArgumentError:
			return {'found': False}

		self._copy_folder_contents(comp.output_folder, destdir)
		return {'found': True}

	def check(self, depinfo, deptile, depsettings):
		from iotile.core.dev.registry import ComponentRegistry

		reg = ComponentRegistry()

		try:
			comp = reg.find_component(depinfo['name'])
		except ArgumentError:
			return True

		if comp.release_date is not None and comp.release_date > deptile.release_date:
			return False

		return True




from iotile.core.utilities.typedargs import context, annotated, param, return_type, iprint
from iotile.core.dev.iotileobj import IOTile
from iotile.core.exceptions import *
from resolverchain import DependencyResolverChain
import os

@context("DependencyManager")
class DependencyManager (object):
	"""Tools to manage IOTile dependencies and build infrastructure.

	"""

	def __init__(self):
		pass

	@param("path", "path", "exists", desc="Path to IOTile to check")
	@return_type('basic_dict')
	def info(self, path="."):
		"""Get information on an IOTile component.
		
		If path is not given, the current directory is assumed to be an IOTile component.
		"""

		tile = IOTile(path)

		info = {
			'is_development_version': not tile.release,
			'dependencies': tile.dependencies
		}

		return info

	@return_type('map(string, string)')
	@param("path", "path", "exists", desc="Path to IOTile to check")
	def list(self, path='.'):
		"""Check if all necessary dependencies of this tile are satisfied

		Returns
		=======
	
		A map with a string value for each dependency where the value is one of:
		- 'not installed' when there is no dependency currently in build/deps
		- 'installed' when the dependency in build/deps is valid
		- 'invalid version' when the dependency in build/deps has an invalid version
		"""

		tile = IOTile(path)

		if tile.release:
			raise ArgumentError("Cannot check dependencies on a release mode tile that cannot have dependencies")

		dep_stati = {}

		for dep in tile.dependencies:
			try:
				deptile = IOTile(os.path.join(path, 'build', 'deps', dep['unique_id']))
			except EnvironmentError, IOError:
				dep_stati[dep['name']] = 'not installed'
				continue

			dep_stati[dep['name']] = 'installed'

			#TODO: Check if the dependencies have the correct version

		return dep_stati

	@return_type('map(string, string)')
	@param("path", "path", "exists", desc="Path to IOTile to check")
	def versions(self, path='.'):
		"""Return the version of all installed dependencies

		Returns
		=======
	
		A map with a string value for each dependency where the value is one of:
		- 'not installed' when there is no dependency currently in build/deps
		- 'X.Y.Z' with the version of the dependency when it is installed
		"""

		tile = IOTile(path)

		if tile.release:
			raise ArgumentError("Cannot check dependencies on a release mode tile that cannot have dependencies")

		dep_stati = {}

		for dep in tile.dependencies:
			try:
				deptile = IOTile(os.path.join(path, 'build', 'deps', dep['unique_id']))
			except EnvironmentError, IOError:
				dep_stati[dep['name']] = 'not installed'
				continue

			dep_stati[dep['name']] = str(deptile.version)

			#TODO: Check if the dependencies have the correct version

		return dep_stati

	@param("path", "path", "exists", desc="Path to IOTile to check")
	def update(self, path='.'):
		"""Attempt to resolve all dependencies in this IOTile by installing them into build/deps
		"""

		tile = IOTile(path)
		if tile.release:
			raise ArgumentError("Cannot update dependencies on a release mode tile that cannot have dependencies")

		depdir = os.path.join(tile.folder, 'build', 'deps')

		#FIXME: Read resolver_settings.json file
		resolver_chain = DependencyResolverChain()
		
		for dep in tile.dependencies:
			result = resolver_chain.update_dependency(tile, dep)
			iprint("Resolving %s: %s" % (dep['name'], result))

	@param("path", "path", "exists", desc="Path to IOTile to check")
	def clean(self, path='.'):
		"""Remove all dependencies of this IOTile from build/deps
		"""

		tile = IOTile(path)
		if tile.release:
			raise ArgumentError("Cannot update dependencies on a release mode tile that cannot have dependencies")

		depdir = os.path.join(tile.folder, 'build', 'deps')

		import shutil
		shutil.rmtree(depdir)
		os.makedirs(depdir)



from docutils.parsers.rst import Directive
import docutils
import os.path
from iotile.core.dev.registry import ComponentRegistry 
import json
import sphinx
from docutils.statemachine import ViewList

#TODO: Modify this to not list whether the dependency is installed or not

class DependenciesDirective (Directive):
	"""Directive for adding a list of all dependencies used by this component

	This directive parses the module_settings.json file for the component and 
	adds links to any installed dependencies that it finds there.
	"""

	def run(self):
		component = os.path.abspath('.')
		deps = self._get_dependencies(component)
		reg = ComponentRegistry()

		found = []
		not_found = []
		
		for name in deps.iterkeys():
			try:
				tile = reg.find_component(name)
			except ArgumentError:
				not_found.append(name)

			found.append(tile)

		deplines = []

		for tile in found:
			deplines.append('- %s' % tile.name)

		for name in not_found:
			deplines.append('- %s (NOT INSTALLED)' % name)

		view = ViewList(deplines, 'dependencies-directive')

		node = docutils.nodes.paragraph()
		sphinx.util.nodes.nested_parse_with_titles(self.state, view, node)
		return node.children

	def _get_dependencies(self, component):		
		module_settings = os.path.join(component, 'module_settings.json')

		with open(module_settings, "r") as f:
			settings = json.load(f)

		mod = settings['modules'].keys()[0]
		mod_settings = settings['modules'][mod]

		if 'depends' not in mod_settings:
			return {}

		return mod_settings['depends']


# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#block.py
#An object representing a TBBlock

from handler import TBHandler
import sys
import os.path
from iotile.build.utilities import template
from iotile.build.build import build
from iotile.core.utilities import intelhex
from iotile.core.exceptions import *
from pkg_resources import resource_filename, Requirement

block_size = 16

hw_offset = 0
major_api_offset = 1
minor_api_offset = 2
name_offset = 3
major_version_offset = 9
minor_version_offset = 10
patch_version_offset = 11
checksum_offset = 12
magic_offset = 13
app_information_offset = 14
reserved = 15

# Indices into the app_information_table
cmd_map_index = 0
interface_list_index = 1
config_list_index = 2
reserved_index = 3

known_hwtypes = {
	10: "NXP LPC824 (Cortex M0+)"
}

class TBBlock:
	"""
	The block in program memory describing a MoMo application module.  The MIB block
	contains information on the application module and a sparse matrix representation
	of a jump table containing all of the command ids and interfaces that the module 
	knows how to respond to.
	"""

	TemplateName 				= 'command_map.asm'
	CTemplateName 				= 'command_map_c.c'
	CTemplateNameHeader 		= 'command_map_c.h'
	ConfigTemplateName			= 'config_variables_c.c'
	ConfigTemplateNameHeader	= 'config_variables_c.h'
	
	def __init__(self):
		"""
		Given an intelhex object, extract the MIB block information
		from it or raise an exception if a TBBlock cannot be found
		at the right location.
		"""

		self.commands = {}
		self.configs = {}
		self.interfaces = []

		self.valid = True
		self.error_msg = ""
		self.hw_type = -1

		self._parse_hwtype()

	@classmethod
	def ParseHardwareType(cls, ih):
		if isinstance(ih, basestring):
			ih = intelhex.IntelHex16bit(ih)
			ih.padding = 0x3FFF

		if ih is None:
			raise ArgumentError("Could not load intelhex file from arguement", argument=strih)

		build_settings = build.load_settings()
		base_addr = build_settings["mib12"]["mib"]["base_address"]
		hw_type = decode_retlw(ih, base_addr + hw_offset)
		
		if hw_type in known_hwtypes:
			return known_hwtypes[hw_type]

		raise DataError("Unknown Hardware Type", hw_type=hw_type)

	def set_api_version(self, major, minor):
		"""
		Set the API version this module was designed for.

		Each module must declare the mib12 API version it was compiled with as a
		2 byte major.minor number.  This information is used by the pic12_executive
		to decide whether the application is compatible.
		"""

		if major > 255 or major < 0 or minor > 255 or minor < 0:
			raise ArgumentError("Invalid API version number with component that does not fit in 1 byte", major=major, minor=minor)
		
		self.api_version = (major, minor)

	def set_module_version(self, major, minor, patch):
		"""
		Set the module version for this module.

		Each module must declare a semantic version number in the form:
		major.minor.patch

		where each component is a 1 byte number between 0 and 255.
		"""

		if major > 255 or major < 0 or minor > 255 or minor < 0 or patch > 255 or patch < 0:
			raise ArgumentError("Invalid module version number with component that does not fit in 1 byte", major=major, minor=minor, patch=patch)

		self.module_version = (major, minor, patch)

	def set_name(self, name):
		"""
		Set the module name to a 6 byte string

		If the string is too short it is appended with space characters.
		"""

		if len(name) > 6:
			raise ArgumentError("Name must be at most 6 characters long", name=name)

		if len(name) < 6:
			name += ' '*(6 - len(name))

		self.name = name

	def add_interface(self, interface):
		"""
		Add an interface to the MIB Block.

		The interface must be a number between 0 and 2^32 - 1, i.e. a non-negative 4 byte integer 
		"""

		if interface < 0 or interface >= 2**32:
			raise ArgumentError("Interface ID is not a non-negative 4-byte number", interface=interface)

		if interface in self.interfaces:
			raise ArgumentError("Attempted to add the same interface twice.", interface=interface, existing_interfaces=self.interfaces)

		self.interfaces.append(interface)

	def add_command(self, cmd_id, handler):
		"""
		Add a command to the TBBlock.  

		The cmd_id must be a non-negative 2 byte number.
		handler should be the command handler
		"""

		if cmd_id < 0 or cmd_id >= 2**16:
			raise ArgumentError("Command ID in mib block is not a non-negative 2-byte number", cmd_id=cmd_id, handler=handler)

		if cmd_id in self.commands:
			raise ArgumentError("Attempted to add the same command ID twice.", cmd_id=cmd_id, existing_handler=self.commands[cmd_id],
								new_handler=handler)

		self.commands[cmd_id] = handler

	def add_config(self, config_id, config_data):
		"""
		Add a configuration variable to the MIB block
		"""

		if config_id < 0 or config_id >= 2**16:
			raise ArgumentError("Config ID in mib block is not a non-negative 2-byte number", config_data=config_id, data=config_data)

		if config_id in self.configs:
			raise ArgumentError("Attempted to add the same command ID twice.", config_data=config_id, old_data=self.configs[config_id],
								new_data=config_data)

		self.configs[config_id] = config_data

	def _check_magic(self, ih):
		magic_addr = self.base_addr + magic_offset
		instr = ih[magic_addr]

		#Magic Value should be a retlw 0xAA
		if instr == 0x34AA:
			return True

		return False

	def _convert_program_address(self, addr):
		"""
		Make sure that addr points to program memory (by having the high bit set) 
		and then strip the high bit
		"""

		if not addr & (1 << 15):
			raise ValueError("Address 0x%X did not have the high bit set indicating a program memory address." % addr)

		return addr & ~(1 << 15)

	def _calculate_app_checksum(self, ih):
		"""
		Calculate the checksum of the application module.

		Automatically find the correct start and stop rows based on the information for this
		hardware type.
		"""

		proc = MIB12Processor.FromChip(self.chip_name)

		start,stop = proc.app_rom

		check = 0
		for i in xrange(start, stop+1):
			low = ih[i] & 0xFF
			high = ih[i] >> 8
			high = high & 0b00111111
			check += low + high

		check = check & 0xFF
		return check

	def _parse_hwtype(self):
		"""
		Convert the numerical hw id to a chip name using the well-known
		conversion table
		"""

		if self.hw_type not in known_hwtypes:
			self.chip_name = "Unknown Chip (type=%d)" % self.hw_type
			return

		self.chip_name = known_hwtypes[self.hw_type]

	def create_c(self, folder):
		"""
		Create a C file that contains a map of all of the mib commands defined in this block

		Also create config files containing definitions for all of the required config variables. 
		"""

		temp = template.RecursiveTemplate(TBBlock.CTemplateName, resource_filename(Requirement.parse("iotile-build"), "iotile/build/config/templates"))
		temp.add(self)
		temp.render(folder)

		temp = template.RecursiveTemplate(TBBlock.ConfigTemplateName, resource_filename(Requirement.parse("iotile-build"), "iotile/build/config/templates"))
		temp.add(self)
		temp.render(folder)

		temp = template.RecursiveTemplate(TBBlock.CTemplateNameHeader, resource_filename(Requirement.parse("iotile-build"), "iotile/build/config/templates"))
		temp.add(self)
		temp.render(folder)

		self.create_config_headers(folder)

	def create_config_headers(self, folder):
		"""
		Create C headers for config variables defined in this block
		"""

		temp = template.RecursiveTemplate(TBBlock.ConfigTemplateNameHeader, resource_filename(Requirement.parse("iotile-build"), "iotile/build/config/templates"))
		temp.add(self)
		temp.render(folder)

	def __str__(self):
		rep  = "MIB Block\n"
		rep += "---------\n"

		if not self.valid:
			rep += "Block invalid: %s\n" % self.error_msg
			return rep

		rep += "Block Valid\n"
		rep += "Name: '%s'\n" % self.name
		rep += "Hardware: %s\n" % self.chip_name
		rep += "API Version: %d.%d\n" % (self.api_version[0], self.api_version[1])
		rep += "Module Version: %d.%d.%d\n" % (self.module_version[0], self.module_version[1], self.module_version[2])

		if hasattr(self, 'stored_checksum'):
			rep += "Stored Checksum: 0x%X\n" % self.stored_checksum
			rep += "Checksum Valid: %s" % (self.app_checksum == 0,)
		else:
			rep += "Checksum Valid: Unknown" 

		rep += "\n# Supported Commands #"
		for id, handler in self.commands.iteritems():
			rep += "\n0x%X: %s" % (id, str(handler))

		rep += "\n\n# Supported Interfaces #"
		for id in self.interfaces:
			rep += "\n0x%X" % id

		return rep


from iotile.core.exceptions import *

class DependencyResolver (object):
	"""An object that is capable of finding and installing a dependency

	"""

	def resolve(self, depinfo, destdir):
		"""Attempt to resolve this dependency using a subclass defined method
		
		Params
		======
		depinfo: a dictionary returned by IOTile.dependencies describing the dependency
		destdir: the directory that the dependency should be copied into

		Returns
		=======
		The function returns a dictionary that has required and optional keys.  The required
		keys are:

		found: boolean if this resolver found a matching dependency

		optional keys are:
		stop: 	boolean if this resolver is suggesting that we should stop looking for this dependency.
				This is useful for making a resolver that stops a resolver chain
		info:	a dictionary that contains information that this resolver wants to store with the dependency
				for future reference.
		"""

		raise NotFoundError("DependencyResolver did not implement resolve method")

	def check(self, depinfo, deptile, depsettings):
		"""Check if this dependency is the latest version in a subclass defined way


		Params
		======
		depinfo: a dictionary returned by IOTile.dependencies describing the dependency
		deptile: an IOTile object for the installed dependency
		depsettings: a dictionary that was previously stored with this dependency by resolve

		Returns
		=======
		A boolean with True meaning the dependency is up-to-date or False if it is not.
		"""

		raise NotFoundError("DependencyResolver did not implement check method")

	def _copy_folder_contents(self, source, dest):
		
		import shutil
		shutil.copytree(source, dest)